
/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.stride;

import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.globals.Globals;

class SetupAction extends Action {
	

	/* input - start value for the count */
	private Terminal startValueTerm;
	/* output - ack for the start value */
	private Terminal ackStartValueTerm;
	
	/* input - step size */
	private Terminal stepTerm;
	/* output - ack step size */
	private Terminal ackStepTerm;
	
	/* input - number of steps */
	private Terminal stepCountTerm;
	/* output - ack for stepCount */
	private Terminal ackStepCountTerm;
	
	/* input - signal that the current cycle is done, enable the subsequuent cycle */
	private Terminal intCycleDoneTerm;
	private Terminal intAckCycleDoneTerm;

	/** Function used in producing "ack signals" */
	OutputFunction ack;
	/** produce and ACK, and also trigger the count */
	OutputFunction ackAndTrigger;
	
	/** command used to trigger the count */
	Command startCountCommand;
	
	public SetupAction(String name, Command c) {
		super(name);
		
		/* record the command to trigger the count */
		startCountCommand= c;
		// make the ack function
		ack= new AckFunction();
		// make the AckAndTrigger function
		ackAndTrigger= new AckAndTrigger(startCountCommand);
		
		// associate the ack function with approproate terminals
		outputFunctions.put(StrideDefaults.ACK_START_VALUE, ack);
		outputFunctions.put(StrideDefaults.ACK_STEP, ack);
		outputFunctions.put(StrideDefaults.INT_ACK_CYCLE_DONE_IN, ack);
		// one of the outputs must engage the trigger
		outputFunctions.put(StrideDefaults.ACK_STEP_COUNT, ackAndTrigger);
	}
	

	/**
	 * Attach a terminal.
	 * Attach only terminals that are needed for this action, ignore the rest.
	 * Self check ensures that all needed terminals are attached
	 */
	public void attach(Terminal t) {
		// get the bas ename for check
		String termName= t.getBaseName();
		// remember whether the terminal was of a bad type
		boolean badType= false;
		// remember whether the same name been added twice?
		boolean twice= false;

		if (termName.equals(StrideDefaults.START_VALUE)) { 
			if (startValueTerm == null) {
				if (t instanceof InputTerminal) {
					t.attachAction(this);
					numInputs++;
					startValueTerm= (InputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.ACK_START_VALUE)) { 
			if (ackStartValueTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					ackStartValueTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.STEP)) { 
			if (stepTerm == null) {
				if (t instanceof InputTerminal) {
					t.attachAction(this);
					numInputs++;
					stepTerm= (InputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.ACK_STEP)) { 
			if (ackStepTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					ackStepTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}		
		else if (termName.equals(StrideDefaults.STEP_COUNT)) { 
			if (stepCountTerm == null) {
				if (t instanceof InputTerminal) {
					t.attachAction(this);
					numInputs++;
					stepCountTerm= (InputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.ACK_STEP_COUNT)) { 
			if (ackStepCountTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					ackStepCountTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}		
		else if (termName.equals(StrideDefaults.INT_CYCLE_DONE_IN)) { 
			if (intCycleDoneTerm == null) {
				if (t instanceof InputTerminal) {
					t.attachAction(this);
					numInputs++;
					intCycleDoneTerm= (InputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.INT_ACK_CYCLE_DONE_IN)) { 
			if (intAckCycleDoneTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					intAckCycleDoneTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else {
			// unknown terminal
			director.error("Terminal " 
					+ t.getName() 
					+ " cannot be attached to SetupAction "
					+ getName() + ", id= " + getID()
					+ "; terminal name not recognized");
		}
		
		// check for errors
		if (badType) {
			// the terminal was of a wrong type
			director.fatalError("SetupAction, " + getName() + ", " + getID() +  
					": Terminal " + t.getName() + " is of a bad type");
		}
		if (twice) {
			// this was the second addition of the same name
			director.fatalError("SetupAction, " + getName() + ", " + getID() 
					+ ": Terminal " + t.getName() + " cannot be attached twice");
		}
		
		
	} // attach
	
	
	/**
	 * This action combines producing a vacuous acknowledgment with
	 * triggering the count.
	 * The trigger piggybacks on computing the acknowledgment value.
	 * The action must be attached to one of the output terminals ONLY,
	 * otherwise, multiple triggers will be executed resulting in an error.
	 * @author ib27688
	 *
	 */
	private class AckAndTrigger extends OutputFunction{

		/** Global settings */
		protected Globals globals= Globals.getInstance();
		
		private Command trigger;
		
		private Delay triggerDelay;
		
		public AckAndTrigger(Command c) {
			super("AckAndTrigger");
			// record the trigger command
			trigger= c;
			Integer fDelay= globals.intValue(StrideDefaults.STRIDE_ACK_DELAY);
			functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: StrideDefaults.STRIDE_ACK_DELAY_DEF;

			fDelay= globals.intValue(StrideDefaults.STRIDE_TRIGGER_DELAY);
			triggerDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: StrideDefaults.STRIDE_ACK_DELAY_DEF;
		}
		

		/** before producing the null value, gather the data needed to start the count */
		public Data value() {
			// gather the values
			// TODO what if the value is not valid???
			BeginCountRecord r= new BeginCountRecord();
			// assume all setup data are valid numbers
			r.valid= true;
			try {
				Payload p= (Payload)(startValueTerm.getData());
				Data d= p.data;
				// get the number
				r.startValue= (Integer)d.value;
				// was the number valid?
				r.valid= r.valid && d.valid;
				
				p= (Payload)(stepTerm.getData());
				d= p.data;
				// get the value
				r.stepSize= (Integer)d.value;
				// update validity
				r.valid= r.valid && d.valid;
				
				p= (Payload)(stepCountTerm.getData());
				d= p.data;
				// get the value
				r.numberOfSteps= (Integer)d.value;
				// update the validity
				r.valid= r.valid && d.valid;
			}
			catch (Exception e) {
				// if there was an exception, it was casting
				// mark setup data as invalid
				r.valid= false;
			}
			// Invoke the trigger, pass on the record of values gathered
			// System.out.println("\n\n" + r + "\n\n");
			director.schedule(trigger, r, triggerDelay);
			
			return new Data();
		} // value
				
	}
	
	/** self check - are all the terminals attached, run the self check from superclass */
	public boolean selfCheck() {
		boolean OK= true;
		boolean check;
		check= (startValueTerm != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.START_VALUE);
		}

		check= (ackStartValueTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.ACK_START_VALUE);
		}
		
		check= (stepTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.STEP);
		}

		check= (ackStepTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.ACK_STEP);
		}

		check= (stepCountTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.STEP_COUNT);
		}

		check= (ackStepCountTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.ACK_STEP_COUNT);
		}
	
		check= (intCycleDoneTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.INT_CYCLE_DONE_IN);
		}
		
		check= (intAckCycleDoneTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.INT_ACK_CYCLE_DONE_IN);
		}
		
		// also run the self check on the superclass
		check= super.selfCheck();
		OK= OK && check;
		
		return OK;
	} // selfCheck
	
} // class SetupAction
