/*
 * Copyright ? 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */


package com.sunlabs.archsim.fleet.ships.tokenaccumulator;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.gasp.State;

/**
 * Accumulator SHIP with Pipeline interface.
 *
 * The Accumulator accepts a sequence of numbers via a self resetting DataIn
 * input terminal. Each incoming data is acknowledged on the ackDataIn output.
 * The Accumulator interface behaves as follows:
 *
 *   Upon receiving a valid data on DataIn, the Accumulator increases its
 *   internal sum by the value transmitted in the data.
 *
 *   Upon receiving an OOB data on DataIn, the Accumulator produces the cumulative
 *   sum of all received data on the DataOut output and resets the sum to 0.
 *
 *   When an overflow occurs the internal overflow bit is set. Upon receiving the
 *   next OOB data on DataIn an OOB data is produced on DataOut with its value set
 *   to plus infinity.
 *
 *   When an underflow occurs the internal underflow bit is set. Upon receiving
 *   the next OOB data on DataIn an OOB data is produced on DataOut with its value
 *   set to minus infinity.
 *
 * The Accumulator is extended with a Pipeline interface, which allows to
 * synchronize SHIPs to prevent them from flooding the Switch Fabric. A valid
 * token means that it's producer is ready to accept new data.
 * An invalid token designates the end of current transaction.
 *
 * @author mh193489
 */
public class TokenAccumulator extends Component {
    
    /* interface terminals */
    
    InputTerminal dataIn;
    OutputTerminal ackDataIn;
    
    OutputTerminal dataOut;
    InputTerminal ackDataOut;
    
    OutputTerminal tokenIn;
    InputTerminal ackTokenIn;
    
    InputTerminal tokenOut;
    OutputTerminal ackTokenOut;
    
    /* intrnal terminals */
    
    private OutputTerminal intDataOut;
    private InputTerminal ackIntDataOut;
    
    private InputTerminal intDataIn;
    private OutputTerminal ackIntDataIn;
    
    private OutputTerminal intTokenIn;
    private InputTerminal ackIntTokenIn;
    
    private InputTerminal intTokenInDummy;
    private OutputTerminal ackIntTokenInDummy;
    
    /**
     * AccumulatingAction is the core action of teh accumulator. Performs
     * the accumulation or outputs an OOB value, depending on the input data.
     */
    AccumulatingAction accumulatorIn;
    
    /**
     * OutputAction makes sure that the result is produced when and only when there is
     * a token waiting on the TokenOut terminal.
     */
    OutputAction accumulatorOut;
    
    /**
     * In order to curcumvent the restriction of the BranchAction of not
     * allowing ot attach the same terminal more than once.
     */
    PassOnAction accumulatorTokenIn, accumulatorAckTokenIn;
    
    /**
     * Constructor for the Accumulator in case xml script is used and the
     * Accumulator is specified with the parameter attribute
     */
    public TokenAccumulator(String name, String dataString) {
        this(name);        
    } // Accumulator
    
    /**
     * Constructor for the Accumulator. Creates all interface terminals as well
     * as theinternal terminals, actions and states.
     *
     * @param name  name of this Accumulator instance
     */
    public TokenAccumulator(String name) {
        super(name, null);
        
        buildTokenAccumulator();

    } // Accumulator
    
    public TokenAccumulator(String name, Parameters params) {
        super(name, null);
        
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
		
				director.error("Component: " + name + ", id= " + getID() 
						+ ": Attribute " + pName + " with value " + sValue
						+ " is ignored");					
		
			} // for
		} // if params != null
		        
        
        buildTokenAccumulator();

    } // Accumulator

    
    protected void buildTokenAccumulator() {
        /* Create interface terminals */
        
        dataIn= new InputTerminal(TokenAccumulatorDefaults.DATA_IN, this);
        ackDataIn= new OutputTerminal(TokenAccumulatorDefaults.ACK_DATA_IN, this);
        
        dataOut= new OutputTerminal(TokenAccumulatorDefaults.DATA_OUT, this);
        ackDataOut= new InputTerminal(TokenAccumulatorDefaults.ACK_DATA_OUT, this);
        
        tokenIn= new OutputTerminal(TokenAccumulatorDefaults.TOKEN_IN, this);
        ackTokenIn= new InputTerminal(TokenAccumulatorDefaults.ACK_TOKEN_IN, this);
        
        tokenOut= new InputTerminal(TokenAccumulatorDefaults.TOKEN_OUT, this);
        ackTokenOut= new OutputTerminal(TokenAccumulatorDefaults.ACK_TOKEN_OUT, this);
        
        /* create internal terminals */
        
        intDataOut= new OutputTerminal(TokenAccumulatorDefaults.INTERNAL_DATA_OUT, this);
        ackIntDataOut= new InputTerminal(TokenAccumulatorDefaults.ACK_INTERNAL_DATA_OUT, this);
        
        intDataIn= new InputTerminal(TokenAccumulatorDefaults.INTERNAL_DATA_IN, this);
        ackIntDataIn= new OutputTerminal(TokenAccumulatorDefaults.ACK_INTERNAL_DATA_IN, this);
        
        intTokenIn= new OutputTerminal(TokenAccumulatorDefaults.INTERNAL_TOKEN_IN, this);
        ackIntTokenIn= new InputTerminal(TokenAccumulatorDefaults.ACK_INTERNAL_TOKEN_IN, this);
        
        intTokenInDummy= new InputTerminal(TokenAccumulatorDefaults.INTERNAL_TOKEN_IN_DUMMY, this);
        ackIntTokenInDummy= new OutputTerminal(TokenAccumulatorDefaults.ACK_INTERNAL_TOKEN_IN_DUMMY, this);
        
        /* create the accumulating action */
        
        accumulatorIn= new AccumulatingAction(TokenAccumulatorDefaults.ACCUMULATING_ACTION);
        this.addMember(accumulatorIn);
        accumulatorIn.attachIn(dataIn, ackDataIn);
        accumulatorIn.attachOut(intDataOut, ackIntDataOut, AccumulatingAction.BRANCH_OUT);
        accumulatorIn.attachOut(intTokenIn, ackIntTokenIn, AccumulatingAction.BRANCH_ACK);
        
        /* create the output action */
        
        accumulatorOut=  new OutputAction();
        this.addMember(accumulatorOut);
        accumulatorOut.attach(intDataIn);
        accumulatorOut.attach(dataOut);
        accumulatorOut.attach(ackDataOut);
        accumulatorOut.attach(tokenOut);
        accumulatorOut.attach(ackTokenOut);
        accumulatorOut.attach(ackIntDataIn);
        
        /* create the passon actions */
        
        accumulatorTokenIn= new PassOnAction("accumulatorTokenIn");
        this.addMember(accumulatorTokenIn);
        accumulatorTokenIn.attach(tokenIn);
        accumulatorTokenIn.attach(intTokenInDummy);
        
        accumulatorAckTokenIn= new PassOnAction("accumulatorAckTokenIn");
        this.addMember(accumulatorAckTokenIn);
        accumulatorAckTokenIn.attach(ackTokenIn);
        accumulatorAckTokenIn.attach(ackIntTokenInDummy);
        
        /* create internal states */
        
        State intDataState= new State(TokenAccumulatorDefaults.INTERNAL_DATA_STATE, this);
        State ackIntDataState= new State(TokenAccumulatorDefaults.ACK_INTERNAL_DATA_STATE, this);
        State intTokenState= new State(TokenAccumulatorDefaults.INTERNAL_TOKEN_STATE, this);
        
        /* make connections */
        
        intDataState.attach(intDataIn);
        intDataState.attach(intDataOut);
        
        ackIntDataState.attach(ackIntDataOut);
        ackIntDataState.attach(ackIntTokenIn);
        ackIntDataState.attach(ackIntTokenInDummy);
        
        intTokenState.attach(intTokenIn);
        intTokenState.attach(ackIntDataIn);
        intTokenState.attach(intTokenInDummy);
        
        
        /* initialize states
         *
         * Note that the ackTokenIn input terminal will we set to active automatically.
         * It will then allow the attached PassOnAction to fire and activate the acks
         * in the AccumulatingAction. Therefore no initialization of internal states required.
         */
    	
    }
        
    /**
     * Get a terminal by name.
     * Return null if no such terminal exists.
     */
    public Terminal getTerminal(String name) {
        if (name.equals(dataIn.getBaseName())
        || name.equals(dataIn.getBaseAlias()))
            return dataIn;
        else if (name.equals(ackDataIn.getBaseName())
        || name.equals(ackDataIn.getBaseAlias()))
            return ackDataIn;
        else if (name.equals(dataOut.getBaseName())
        || name.equals(dataOut.getBaseAlias()))
            return dataOut;
        else if (name.equals(ackDataOut.getBaseName())
        || name.equals(ackDataOut.getBaseAlias()))
            return ackDataOut;
        
        else if (name.equals(tokenIn.getBaseName())
        || name.equals(tokenIn.getBaseAlias()))
            return tokenIn;
        else if (name.equals(ackTokenIn.getBaseName())
        || name.equals(ackTokenIn.getBaseAlias()))
            return ackTokenIn;
        else if (name.equals(tokenOut.getBaseName())
        || name.equals(tokenOut.getBaseAlias()))
            return tokenOut;
        else if (name.equals(ackTokenOut.getBaseName())
        || name.equals(ackTokenOut.getBaseAlias()))
            return ackTokenOut;
        else
            return null;
        
    } // getTerminal
    
} // TestShip;


